BASE_GRAPH_URL = 'https://graph.facebook.com'
BASE_AUTHORIZATION_URL = 'https://www.facebook.com'

GRAPH_VERSION_V2_5 = 'v2.5'
GRAPH_VERSION_V2_6 = 'v2.6'
GRAPH_VERSION_V2_7 = 'v2.7'
GRAPH_VERSION_V2_8 = 'v2.8'
GRAPH_VERSION_V2_9 = 'v2.9'
GRAPH_VERSION_V2_10 = 'v2.10'
GRAPH_VERSION_V2_11 = 'v2.11'
GRAPH_VERSION_V2_12 = 'v2.12'

ISO_8601_DATE_FORMAT = '%Y-%m-%dT%H:%M:%S%z'

DEFAULT_GRAPH_VERSION = GRAPH_VERSION_V2_12
DEFAULT_REQUEST_TIMEOUT = 60

METHOD_DELETE = 'DELETE'
METHOD_GET = 'GET'
METHOD_POST = 'POST'
METHOD_PUT = 'PUT'
